@@ -20,9 +20,7 @@ from math import ceil
 import re
 from xlrd.biffh import XLRDError
 
-with warnings.catch_warnings():
-    warnings.simplefilter("ignore", category=UserWarning)
-    from thefuzz import fuzz
+from rapidfuzz import fuzz
 
 try:
     import geopandas as gpd
@@ -111,7 +109,7 @@ class Data_Loader(ABC):
         pass
 
     @abstractmethod
-    def load(self, year=None, nrows=None, *, pbar=True, agency=None, opt_filter=None, select=None, output_type=None):
+    def load(self, year=None, nrows=None, offset=0, *, pbar=True, agency=None, opt_filter=None, select=None, output_type=None):
         pass
 
     def get_years(self, *, nrows=1, **kwargs):
@@ -139,9 +137,9 @@ class Data_Loader(ABC):
         misses = 0
         years = []
         while misses < max_misses:
-            df = self.load(year=year, nrows=nrows)
+            count = self.get_count(year=year)
 
-            if not hasattr(df, '__len__') or len(df)==0:  # If doesn't have len attribute, it is None
+            if count==0:  # If doesn't have len attribute, it is None
                 misses+=1
             else:
                 misses = 0
@@ -170,7 +168,7 @@ class Csv(Data_Loader):
 
     Methods
     -------
-    load(year=None, nrows=None, pbar=True, agency=None)
+    load(year=None, nrows=None, offset=0, pbar=True, agency=None)
         Load data for query
     get_count(year=None, agency=None, force=False)
         Get number of records/rows generated by query
@@ -237,7 +235,7 @@ class Csv(Data_Loader):
                 "efficient. If running get_count with a year argument is still desired, set force=True")
 
 
-    def load(self, year=None, nrows=None, *, pbar=True, agency=None, **kwargs):
+    def load(self, year=None, nrows=None, offset=0, *, pbar=True, agency=None, **kwargs):
         '''Download CSV file to pandas DataFrame
         
         Parameters
@@ -246,6 +244,8 @@ class Csv(Data_Loader):
             (Optional) Either the year or the year range [first_year, last_year] for the data that is being requested. None value returns data for all years.
         nrows : int
             (Optional) Only returns the first nrows rows of the CSV
+        offset - int
+            (Optional) Number of records to offset from first record. Default is 0 to return records starting from the first.
         pbar : bool
             (Optional) If true (default), a progress bar will be displayed
         agency : str
@@ -280,12 +280,18 @@ class Csv(Data_Loader):
                 raise e
             with requests.get(self.url, params=None, stream=True) as resp:
                 try:
-                    table = pd.read_csv(TqdmReader(resp, pbar=pbar), nrows=nrows, encoding_errors='surrogateescape')
+                    table = pd.read_csv(TqdmReader(resp, pbar=pbar), nrows=offset+nrows if nrows is not None else None, encoding_errors='surrogateescape')
                 except (urllib.error.HTTPError, pd.errors.ParserError) as e:
                     raise OPD_DataUnavailableError(*e.args, _url_error_msg.format(self.url))
                 except Exception as e:
                     raise e
 
+        if offset>0:
+            rows_limit = offset+nrows if nrows is not None and offset+nrows<len(table) else len(table)
+            table = table.iloc[offset:rows_limit].reset_index(drop=True)
+        if nrows is not None and len(table)>nrows:
+            table = table.head(nrows)
+
         table = filter_dataframe(table, date_field=self.date_field, year_filter=year, 
             agency_field=self.agency_field, agency=agency)
 
@@ -335,7 +341,7 @@ class Excel(Data_Loader):
 
     Methods
     -------
-    load(year=None, nrows=None, pbar=True, agency=None)
+    load(year=None, nrows=None, offset=0, pbar=True, agency=None)
         Load data for query
     get_count(year=None, agency=None, force=False)
         Get number of records/rows generated by query
@@ -485,7 +491,7 @@ class Excel(Data_Loader):
         return names, False
 
 
-    def load(self, year=None, nrows=None, *, agency=None, **kwargs):
+    def load(self, year=None, nrows=None, offset=0, *, agency=None, **kwargs):
         '''Download Excel file to pandas DataFrame
         
         Parameters
@@ -494,6 +500,8 @@ class Excel(Data_Loader):
             (Optional) Either the year or the year range [first_year, last_year] for the data that is being requested. None value returns data for all years.
         nrows : int
             (Optional) Only returns the first nrows rows of the Excel
+        offset - int
+            (Optional) Number of records to offset from first record. Default is 0 to return records starting from the first.
         agency : str
             (Optional) Name of the agency to filter for. None value returns data for all agencies.
             
@@ -505,6 +513,7 @@ class Excel(Data_Loader):
         Note: Older Excel files (.xls) and OpenDocument file formats (.odf, .ods, .odt) are not supported. Please submit an issue if this is needed.
         '''
 
+        nrows_read = offset+nrows if nrows is not None else None
         year_dict, has_year_sheets = self.__get_sheets()
 
         with warnings.catch_warnings():        
@@ -522,7 +531,7 @@ class Excel(Data_Loader):
                 cols_added = 0
                 for y in range(year[0], year[1]+1):
                     if y in year_dict:
-                        df = pd.read_excel(self.excel_file, nrows=nrows, sheet_name=year_dict[y])
+                        df = pd.read_excel(self.excel_file, nrows=nrows_read, sheet_name=year_dict[y])
 
                         df = self.__clean(df)
 
@@ -555,16 +564,19 @@ class Excel(Data_Loader):
                                     raise ValueError("Columns don't match")
                             table = pd.concat([table, df], ignore_index=True)
 
-                        if nrows!=None and len(table)>=nrows:
+                        if nrows_read!=None and len(table)>=nrows_read:
                             break
 
                 if isinstance(table, type(None)):
                     return table
             else:
-                table = pd.read_excel(self.excel_file, nrows=nrows)
+                table = pd.read_excel(self.excel_file, nrows=nrows_read)
                 table = self.__clean(table)               
 
-        if nrows!=None and len(table) > nrows:
+        if offset>0:
+            rows_limit = nrows_read if nrows_read is not None and nrows_read<len(table) else len(table)
+            table = table.iloc[offset:rows_limit].reset_index(drop=True)
+        if nrows is not None and len(table)>nrows:
             table = table.head(nrows)
 
         # Clean up column names
@@ -649,7 +661,7 @@ class Arcgis(Data_Loader):
 
     Methods
     -------
-    load(year=None, nrows=None, pbar=True)
+    load(year=None, nrows=None, offset=0, pbar=True)
         Load data for query
     get_count(year=None, where=None)
         Get number of records/rows generated by query
@@ -813,6 +825,7 @@ class Arcgis(Data_Loader):
                 params["resultRecordCount"] = count
 
         params["f"] = out_type
+        params["cacheHint"] = False
 
         try:
             r = requests.get(url, params=params)
@@ -862,45 +875,55 @@ class Arcgis(Data_Loader):
     def _build_date_query(self, year):
 
         # List of error messages that can occur for bad queries as we search for the right query format
-        query_err_msg = ["Unable to complete operation", "Failed to execute query", "Unable to perform query", "Database error has occurred"]
+        query_err_msg = ["Unable to complete operation", "Failed to execute query", "Unable to perform query", "Database error has occurred", 
+                         "'where' parameter is invalid", "Parsing error"]
         
         where_query = ""
         zero_found = False
-        if self._date_format==0 or self._date_format==None:
+        if self._date_format in [0,1] or self._date_format==None:
             start_date, stop_date = _process_date(year)
             
-            where_query = f"{self.date_field} >= '{start_date}' AND  {self.date_field} < '{stop_date}'"
-        
-            try:
-                record_count = self.__request(where=where_query, return_count=True)
-                if "count" not in record_count:
-                    err = "Error Code {}: ".format(record_count["error"]["code"])
-                    if len(record_count["error"]["message"])!=0:
-                        err+=record_count["error"]["message"]
-                        err+=" "
-                    if len(record_count["error"]["details"])>0 and len(record_count["error"]["details"][0])>0:
-                        err+=record_count["error"]["details"][0]
-                    raise KeyError(err)
-                record_count = record_count["count"]
-                if self.verify:
-                    record_count_orig = self.__active_layer.query(where=where_query, return_count_only=True)
-                    if record_count_orig!=record_count:
-                        raise ValueError(f"Record count of {record_count} does not equal count from arcgis package of {record_count_orig}")
-                if self._date_format!=None or record_count>0:
-                    self._date_format = 0
-                    return where_query, record_count
-                else:
-                    zero_found = True
-            except Exception as e:
-                if len(e.args)>0 and "Error Code: 429" in e.args[0]:
-                    raise OPD_TooManyRequestsError(self.url, *e.args, _url_error_msg.format(self.url))
-                elif len(e.args)>0 and any([x in e.args[0] for x in query_err_msg]):
-                    # This query throws an error for this dataset. Try another one below
-                    pass
+            for k in range(0,2):
+                if k==0:
+                    where_query = f"{self.date_field} >= '{start_date}' AND  {self.date_field} < '{stop_date}'"
                 else:
+                    break
+                    # Dataset (San Jose crash data) that required this does not function well so removing its functionality for now to speed up this function.
+                    # This is the recommended way but it has been found to not work sometimes. One dataset was found that requires this.
+                    # https://gis.stackexchange.com/questions/451107/arcgis-rest-api-unable-to-complete-operation-on-esrifieldtypedate-in-query
+                    stop_date_tmp = stop_date.replace("T"," ")
+                    where_query = f"{self.date_field} >= TIMESTAMP '{start_date}' AND  {self.date_field} < TIMESTAMP '{stop_date_tmp}'"
+            
+                try:
+                    record_count = self.__request(where=where_query, return_count=True)
+                    if "count" not in record_count:
+                        err = "Error Code {}: ".format(record_count["error"]["code"])
+                        if len(record_count["error"]["message"])!=0:
+                            err+=record_count["error"]["message"]
+                            err+=" "
+                        if len(record_count["error"]["details"])>0 and len(record_count["error"]["details"][0])>0:
+                            err+=record_count["error"]["details"][0]
+                        raise KeyError(err)
+                    record_count = record_count["count"]
+                    if self.verify:
+                        record_count_orig = self.__active_layer.query(where=where_query, return_count_only=True)
+                        if record_count_orig!=record_count:
+                            raise ValueError(f"Record count of {record_count} does not equal count from arcgis package of {record_count_orig}")
+                    if self._date_format!=None or record_count>0:
+                        self._date_format = k
+                        return where_query, record_count
+                    else:
+                        zero_found = True
+                except Exception as e:
+                    if len(e.args)>0 and "Error Code: 429" in e.args[0]:
+                        raise OPD_TooManyRequestsError(self.url, *e.args, _url_error_msg.format(self.url))
+                    elif len(e.args)>0 and any([x in e.args[0] for x in query_err_msg]):
+                        # This query throws an error for this dataset. Try another one below
+                        pass
+                    else:
+                        raise
+                except:
                     raise
-            except:
-                raise
 
 
         where_formats = [
@@ -912,7 +935,7 @@ class Arcgis(Data_Loader):
         # Make year iterable
         year = [year] if isinstance(year, numbers.Number) else year
 
-        if self._date_format not in [None, 0] and any([isinstance(x,str) and len(x)!=4 for x in year]):
+        if self._date_format not in [None, 0, 1] and any([isinstance(x,str) and len(x)!=4 for x in year]):
             # Currently can only handle years
             raise ValueError("Currently unable to handle non-year inputs")
 
@@ -959,7 +982,7 @@ class Arcgis(Data_Loader):
         return "", 0
 
     
-    def load(self, year=None, nrows=None, *, pbar=True, **kwargs):
+    def load(self, year=None, nrows=None, offset=0, *, pbar=True, **kwargs):
         '''Download table from ArcGIS to pandas or geopandas DataFrame
         
         Parameters
@@ -968,6 +991,8 @@ class Arcgis(Data_Loader):
             (Optional) Either the year or the year range [first_year, last_year] for the data that is being requested.  None value returns data for all years.
         nrows : int
             (Optional) Only returns the first nrows rows of the table
+        offset - int
+            (Optional) Number of records to offset from first record. Default is 0 to return records starting from the first.
         pbar : bool
             (Optional) If true (default), a progress bar will be displayed
             
@@ -978,8 +1003,10 @@ class Arcgis(Data_Loader):
         '''
         
         where_query, record_count = self.__construct_where(year)
-
-        if record_count==0:
+        
+        # Update record count for request record offset
+        record_count-=offset
+        if record_count<=0:
             return None
 
         batch_size = self.max_record_count or _default_limit
@@ -993,8 +1020,9 @@ class Arcgis(Data_Loader):
             
         features = []
         for batch in range(num_batches):
+            bs = batch_size if batch<num_batches-1 else nrows-batch*batch_size
             try:
-                data = self.__request(where=where_query, offset=batch*batch_size, count=batch_size)
+                data = self.__request(where=where_query, offset=offset+batch*batch_size, count=bs)
                 features.extend(data["features"])
                 if self.verify:
                     layer_query_result_old = self.__active_layer.query(where=where_query, result_offset=batch*batch_size, 
@@ -1114,7 +1142,7 @@ class Carto(Data_Loader):
 
     Methods
     -------
-    load(year=None, nrows=None, pbar=True)
+    load(year=None, nrows=None, offset=0, pbar=True)
         Load data for query
     get_count(year=None, where=None)
         Get number of records/rows generated by query
@@ -1222,7 +1250,7 @@ class Carto(Data_Loader):
         return where_query
 
     
-    def load(self, year=None, nrows=None, *, pbar=True, **kwargs):
+    def load(self, year=None, nrows=None, offset=0, *, pbar=True, **kwargs):
         '''Download table to pandas or geopandas DataFrame
         
         Parameters
@@ -1231,6 +1259,8 @@ class Carto(Data_Loader):
             (Optional) Either the year or the year range [first_year, last_year] for the data that is being requested.  None value returns data for all years.
         nrows : int
             (Optional) Only returns the first nrows rows of the table
+        offset - int
+            (Optional) Number of records to offset from first record. Default is 0 to return records starting from the first.
         pbar : bool
             (Optional) If true (default), a progress bar will be displayed
             
@@ -1244,7 +1274,8 @@ class Carto(Data_Loader):
         json = self.__request(where=where_query, return_count=True)
         record_count = json["rows"][0]["count"]
 
-        if record_count==0:
+        record_count-=offset
+        if record_count<=0:
             return None
 
         batch_size = _default_limit
@@ -1261,8 +1292,10 @@ class Carto(Data_Loader):
             
         features = []
         for batch in range(num_batches):
+            bs = batch_size if batch<num_batches-1 else nrows-batch*batch_size
+
             try:
-                data = self.__request(where=where_query, offset=batch*batch_size, count=batch_size)
+                data = self.__request(where=where_query, offset=offset+batch*batch_size, count=bs)
                 features.extend(data["features"])
 
                 if batch==0 and len(features)>0:
@@ -1337,7 +1370,7 @@ class Socrata(Data_Loader):
 
     Methods
     -------
-    load(year=None, nrows=None, pbar=True, opt_filter=None, select=None, output_type=None)
+    load(year=None, nrows=None, offset=0, pbar=True, opt_filter=None, select=None, output_type=None)
         Load data for query
     get_count(year=None, opt_filter=None, where=None)
         Get number of records/rows generated by query
@@ -1414,10 +1447,10 @@ class Socrata(Data_Loader):
         except:
             num_rows = float(results[0]["count_1"]) # Value used in VT Shootings data
 
-        return num_rows
+        return int(num_rows)
 
 
-    def load(self, year=None, nrows=None, *, pbar=True, opt_filter=None, select=None, output_type=None, **kwargs):
+    def load(self, year=None, nrows=None, offset=0, *, pbar=True, opt_filter=None, select=None, output_type=None, **kwargs):
         '''Download table from Socrata to pandas or geopandas DataFrame
         
         Parameters
@@ -1426,6 +1459,8 @@ class Socrata(Data_Loader):
             (Optional) Either the year or the year range [first_year, last_year] for the data that is being requested.  None value returns data for all years.
         nrows : int
             (Optional) Only returns the first nrows rows of the table
+        offset - int
+            (Optional) Number of records to offset from first record. Default is 0 to return records starting from the first.
         pbar : bool
             (Optional) If true (default), a progress bar will be displayed
         opt_filter : str
@@ -1442,7 +1477,7 @@ class Socrata(Data_Loader):
         '''
 
         N = 1  # Initialize to value > 0 so while loop runs
-        offset = 0
+        start_offset = offset
 
         where = self.__construct_where(year, opt_filter)
         
@@ -1454,7 +1489,8 @@ class Socrata(Data_Loader):
             use_gpd = _has_gpd
 
         record_count = int(self.get_count(where=where))
-        if record_count==0:
+        record_count-=offset
+        if record_count<=0:
             return None
         batch_size =  _default_limit
         nrows = nrows if nrows!=None and record_count>=nrows else record_count
@@ -1493,7 +1529,7 @@ class Socrata(Data_Loader):
                         break
 
             if output_type == "set":
-                if offset==0:
+                if offset==start_offset:
                     df = set()
 
                 if len(results)>0:
@@ -1503,7 +1539,7 @@ class Socrata(Data_Loader):
                     df.update(results)
 
             elif output_type == "list":
-                if offset==0:
+                if offset==start_offset:
                     df = list()
 
                 if len(results)>0:
@@ -1533,14 +1569,14 @@ class Socrata(Data_Loader):
                 if len(results)>0:
                     new_gdf = gpd.GeoDataFrame.from_features(geojson, crs=4326)
                         
-                    if offset==0:
+                    if offset==start_offset:
                         df = new_gdf
                     else:
                         df = pd.concat([df, new_gdf], ignore_index=True)
             else:
                 output_type = "DataFrame"
                 rows = pd.DataFrame.from_records(results)
-                if offset==0:
+                if offset==start_offset:
                     df = pd.DataFrame(rows)
                 else:
                     df = pd.concat([df, rows], ignore_index=True)
@@ -1556,6 +1592,9 @@ class Socrata(Data_Loader):
 
         if show_pbar:
             bar.close()
+
+        if isinstance(df, pd.DataFrame) and nrows is not None and len(df)>nrows:
+            df = df.head(nrows)
         return df
 
 
