@@ -51,4 +51,10 @@ class OPD_arcgisAuthInfoError(OPD_GeneralError):
     pass
 
 class OPD_SocrataHTTPError(OPD_GeneralError):
+    pass
+
+class OPD_FutureError(OPD_GeneralError):
+    pass
+
+class OPD_MinVersionError(OPD_GeneralError):
     pass
\ No newline at end of file
